import { instantiateWasmModule } from './wasm-helpers.js';
import { proxyWithEngineNameMapping, proxyAsModuleWithBindings, getFloatArrayType } from '../../run/run-helpers.js';
import { createCommonsBindings } from '../../stdlib/commons/bindings-assemblyscript.js';
import { createEngineLifecycleBindings } from './engine-lifecycle-bindings.js';
import { readMetadata } from './metadata.js';
import { ioMsgSendersImports, createIoMessageReceiversBindings, createIoMessageSendersBindings } from './io-bindings.js';
import { createFsImports, createFsBindings } from '../../stdlib/fs/bindings-assemblyscript.js';

/*
 * Copyright (c) 2022-2023 Sébastien Piquemal <sebpiq@protonmail.com>, Chris McCormick.
 *
 * This file is part of WebPd
 * (see https://github.com/sebpiq/WebPd).
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 */
const createEngine = async (wasmBuffer, additionalBindings) => {
    // Create engine context
    // We need to read metadata before everything, because it is used by other initialization functions
    const metadata = await readMetadata(wasmBuffer);
    const bitDepth = metadata.settings.audio.bitDepth;
    const arrayType = getFloatArrayType(bitDepth);
    const engineContext = {
        refs: {},
        metadata: metadata,
        cache: {
            wasmOutput: new arrayType(0),
            wasmInput: new arrayType(0),
            arrayType,
            bitDepth,
            blockSize: 0,
        },
    };
    // Create raw module
    const wasmImports = {
        ...createFsImports(engineContext),
        ...ioMsgSendersImports(engineContext),
    };
    const wasmInstance = await instantiateWasmModule(wasmBuffer, {
        input: wasmImports,
    });
    engineContext.refs.rawModule = proxyWithEngineNameMapping(wasmInstance.exports, metadata.compilation.variableNamesIndex);
    // Create engine
    const engineBindings = createEngineBindings(engineContext);
    const engine = proxyAsModuleWithBindings(engineContext.refs.rawModule, {
        ...engineBindings,
        ...(additionalBindings || {}),
    });
    engineContext.refs.engine = engine;
    return engine;
};
const createEngineBindings = (engineContext) => {
    const { metadata, refs } = engineContext;
    const exportedNames = metadata.compilation.variableNamesIndex.globals;
    // Create bindings for io
    const io = {
        messageReceivers: proxyAsModuleWithBindings(refs.rawModule, createIoMessageReceiversBindings(engineContext)),
        messageSenders: proxyAsModuleWithBindings(refs.rawModule, createIoMessageSendersBindings(engineContext)),
    };
    // Create bindings for core modules
    const globalsBindings = {
        commons: {
            type: 'proxy',
            value: proxyAsModuleWithBindings(refs.rawModule, createCommonsBindings(engineContext)),
        },
    };
    if ('fs' in exportedNames) {
        const fs = proxyAsModuleWithBindings(refs.rawModule, createFsBindings(engineContext));
        globalsBindings.fs = { type: 'proxy', value: fs };
    }
    // Build the full module
    return {
        ...createEngineLifecycleBindings(engineContext),
        metadata: { type: 'proxy', value: metadata },
        globals: {
            type: 'proxy',
            value: proxyAsModuleWithBindings(refs.rawModule, globalsBindings),
        },
        io: { type: 'proxy', value: io },
    };
};

export { createEngine, createEngineBindings };
